@@ -42,7 +42,8 @@ extern "C" {
 # define rhash_aligned_free(ptr) _aligned_free(ptr)
 
 #elif !defined(NO_STDC_ALIGNED_ALLOC) && (__STDC_VERSION__ >= 201112L || defined(_ISOC11_SOURCE)) \
-	&& !defined(__APPLE__) && (!defined(__ANDROID_API__) || __ANDROID_API__ >= 28)
+	&& !defined(__APPLE__) && !defined(__HAIKU__) \
+	&& (!defined(__ANDROID_API__) || __ANDROID_API__ >= 28)
 
 # define HAS_STDC_ALIGNED_ALLOC
 # include <stdlib.h>
