@@ -51,15 +51,23 @@ win32()
   return 1
 }
 
+mingw_or_ucrt()
+{
+  case "$MSYSTEM" in
+    MINGW32|MINGW64|UCRT32|UCRT64) return 0 ;;
+  esac
+  return 1
+}
+
 # detect shared library
 if [ -n "$OPT_SHARED" -a -d "$UPPER_DIR/librhash" ]; then
   D="$UPPER_DIR/librhash"
   N="$D/librhash"
-  if [ -r $N.0.dylib ] && ( uname -s | grep -qi "^darwin" || [ ! -r $N.so.0 ] ); then
+  if [ -r $N.1.dylib ] && ( uname -s | grep -qi "^darwin" || [ ! -r $N.so.1 ] ); then
     export DYLD_LIBRARY_PATH="$D:$DYLD_LIBRARY_PATH"
-  elif ls $D/*rhash.dll 2>/dev/null >/dev/null && ( win32 || [ ! -r $N.so.0 ] ); then
+  elif ls $D/*rhash.dll 2>/dev/null >/dev/null && ( win32 || [ ! -r $N.so.1 ] ); then
     export PATH="$D:$PATH"
-  elif [ -r $N.so.0 ]; then
+  elif [ -r $N.so.1 ]; then
     export LD_LIBRARY_PATH="$D:$LD_LIBRARY_PATH"
   else
     echo "shared library not found at $D"
@@ -113,16 +121,15 @@ new_test() {
 }
 
 print_failed() {
-    st=$( test "$1" = "." -o "$sub_test" -gt 1 && echo " Subtest #$sub_test" )
-    echo "Failed$st"
+    st=$( test "$1" = "." -o "$sub_test" -gt 1 && printf " Subtest #$sub_test" )
+    printf "Failed$st\n"
 }
 
 # verify obtained value $1 against the expected value $2
 check() {
-#printf "test '%s' = '%s'", "$1" "$2"
   sub_test=$((sub_test+1))
   if [ "$1" = "$2" ]; then
-    test "$3" = "." || echo "Ok"
+    test "$3" = "." || printf "Ok\n"
   else
     print_failed "$3"
     printf "obtained: \"%s\"\n" "$1"
@@ -135,7 +142,7 @@ check() {
 
 # match obtained value $1 against given grep-regexp $2
 match_line() {
-  if echo "$1" | grep -vq "$2"; then
+  if printf "$1" | grep -vq "$2"; then
     printf "obtained: \"%s\"\n" "$1"
     printf "regexp:  /%s/\n" "$2"
     fail_cnt=$((fail_cnt+1))
@@ -154,29 +161,29 @@ match() {
     fail_cnt=$((fail_cnt+1))
     return 1;
   else
-    test "$3" = "." || echo "Ok"
+    test "$3" = "." || printf "Ok\n"
   fi
   return 0
 }
 
 new_test "test with a text string:    "
-TEST_RESULT=$( $rhash --message "abc" | tail -1 )
+TEST_RESULT=$( $rhash --message "abc" | tail -n1 )
 TEST_EXPECTED="(message) 352441C2"
 check "$TEST_RESULT" "$TEST_EXPECTED"
 
 new_test "test stdin processing:      "
 TEST_STR="test_string1"
-TEST_RESULT=$( printf "abc" | $rhash -CHMETAGW --sfv - | tail -1 )
+TEST_RESULT=$( printf "abc" | $rhash -CHMETAGW --sfv - | tail -n1 )
 TEST_EXPECTED="(stdin) 352441C2 900150983CD24FB0D6963F7D28E17F72 A9993E364706816ABA3E25717850C26C9CD0D89D ASD4UJSEH5M47PDYB46KBTSQTSGDKLBHYXOMUIA A448017AAF21D8525FC10AE87AA6729D VGMT4NSHA2AWVOR6EVYXQUGCNSONBWE5 4E2448A4C6F486BB16B6562C73B4020BF3043E3A731BCE721AE1B303D97E6D4C7181EEBDB6C57E277D0E34957114CBD6C797FC9D95D8B582D225292076D4EEF5 4E2919CF137ED41EC4FB6270C61826CC4FFFB660341E0AF3688CD0626D23B481"
 check "$TEST_RESULT" "$TEST_EXPECTED"
 
 new_test "test with 1Kb data file:    "
-TEST_RESULT=$( $rhash --printf "%f %C %M %H %E %G %T %A %W\n" test1K.data 2>/dev/null )
-TEST_EXPECTED="test1K.data B70B4C26 B2EA9F7FCEA831A4A63B213F41A8855B 5B00669C480D5CFFBDFA8BDBA99561160F2D1B77 5AE257C47E9BE1243EE32AABE408FB6B 7A6682133082A49C37DB7B008394AEB9C184D5FB2A8D2A6251DD4BBA5F6744B4 4OQY25UN2XHIDQPV5U6BXAZ47INUCYGIBK7LFNI LMAGNHCIBVOP7PP2RPN2TFLBCYHS2G3X D606B7F44BD288759F8869D880D9D4A2F159D739005E72D00F93B814E8C04E657F40C838E4D6F9030A8C9E0308A4E3B450246250243B2F09E09FA5A24761E26B"
+TEST_RESULT=$( $rhash --printf "%f %d %C %M %H %E %G %T %A %W\n" test1K.data 2>/dev/null )
+TEST_EXPECTED="test1K.data . B70B4C26 B2EA9F7FCEA831A4A63B213F41A8855B 5B00669C480D5CFFBDFA8BDBA99561160F2D1B77 5AE257C47E9BE1243EE32AABE408FB6B 7A6682133082A49C37DB7B008394AEB9C184D5FB2A8D2A6251DD4BBA5F6744B4 4OQY25UN2XHIDQPV5U6BXAZ47INUCYGIBK7LFNI LMAGNHCIBVOP7PP2RPN2TFLBCYHS2G3X D606B7F44BD288759F8869D880D9D4A2F159D739005E72D00F93B814E8C04E657F40C838E4D6F9030A8C9E0308A4E3B450246250243B2F09E09FA5A24761E26B"
 check "$TEST_RESULT" "$TEST_EXPECTED" .
 # test calculation/verification of reversed GOST hashes with 1Kb data file
 TEST_RESULT=$( $rhash --simple --gost --gost-cryptopro --gost-reverse test1K.data )
-TEST_EXPECTED="test1K.data  bb4c042bacee51bbabc186107e6020b20991fd4ea119672da24dbe5deeb30b89  06cc52d9a7fb5137d01667d1641683620060391722a56222bb4b14ab332ec9d9"
+TEST_EXPECTED="test1K.data bb4c042bacee51bbabc186107e6020b20991fd4ea119672da24dbe5deeb30b89 06cc52d9a7fb5137d01667d1641683620060391722a56222bb4b14ab332ec9d9"
 check "$TEST_RESULT" "$TEST_EXPECTED" .
 TEST_RESULT=$( $rhash --simple --gost --gost-cryptopro --gost-reverse test1K.data | $rhash -vc - 2>/dev/null | grep test1K.data )
 match "$TEST_RESULT" "^test1K.data *OK"
@@ -207,9 +214,9 @@ $rhash test1K.data | tr -d '\r' | (
   read l; match_line "$l" "^test1K.data B70B4C26\$"
 ) > "$MATCH_LOG"
 if [ ! -s "$MATCH_LOG" ]; then
-  echo "Ok"
+  printf "Ok\n"
 else
-  echo "Failed"
+  printf "Failed\n"
   fail_cnt=$((fail_cnt+1))
   cat "$MATCH_LOG"
 fi
@@ -221,15 +228,36 @@ TEST_EXPECTED="(message) 1 E8B7BE43 5c334qy BTAXLOOA6G3KQMODTHRGS5ZGME hvfkN/qlp
 check "$TEST_RESULT" "$TEST_EXPECTED"
 
 new_test "test %u modifier:           "
-mkdir dir1 && printf "a" > "dir1/=@+.txt"
-TEST_RESULT=$( $rhash -p '%uf %Uf %up %Up %uxc %uxC %ubc %ubC\n' "dir1/=@+.txt" )
-TEST_EXPECTED="%3d%40%2b.txt %3D%40%2B.txt dir1%2f%3d%40%2b.txt dir1%2F%3D%40%2B.txt e8b7be43 E8B7BE43 5c334qy 5C334QY"
+mkdir -p "dir1/d@" && printf "a" > "dir1/d@/=@+.txt"
+TEST_RESULT=$( $rhash -p '%uf %Uf %ud %Ud %Up %uxc\n' "dir1/d@/=@+.txt" )
+TEST_EXPECTED="%3d%40%2b.txt %3D%40%2B.txt dir1%2fd%40 dir1%2Fd%40 dir1%2Fd%40%2F%3D%40%2B.txt e8b7be43"
+check "$TEST_RESULT" "$TEST_EXPECTED" .
+TEST_RESULT=$( $rhash -p '%up %uxc %uxC %ubc %ubC\n' "dir1/d@/=@+.txt" )
+TEST_EXPECTED="dir1%2fd%40%2f%3d%40%2b.txt e8b7be43 E8B7BE43 5c334qy 5C334QY"
 check "$TEST_RESULT" "$TEST_EXPECTED" .
 TEST_RESULT=$( $rhash -p '%uBc %UBc %Bc %u@c %U@c\n' -m "a" )
 TEST_EXPECTED="6Le%2bQw%3d%3d 6Le%2BQw%3D%3D 6Le+Qw== %e8%b7%beC %E8%B7%BEC"
 check "$TEST_RESULT" "$TEST_EXPECTED"
 
 new_test "test special characters:    "
+if ! win32; then
+  NAME_R="$(printf 'a/\r1')"
+  NAME_N="$(printf 'a/\n2')"
+  if mkdir a && touch "$NAME_R" "$NAME_N" 2>/dev/null && test -r "$NAME_R" && test -r "$NAME_N"; then
+    TEST_RESULT=$( $rhash -p '\^%f ' "$NAME_R" "$NAME_N" )
+    TEST_EXPECTED='\\r1 \\n2 '
+    check "$TEST_RESULT" "$TEST_EXPECTED" .
+    TEST_RESULT=$( $rhash -p '\^%p ' "$NAME_R" "$NAME_N" )
+    TEST_EXPECTED='\a/\r1 \a/\n2 '
+    check "$TEST_RESULT" "$TEST_EXPECTED" .
+    TEST_RESULT=$( printf '\\00000000 a/\\r1\n' | $rhash -c --brief - 2>&1 | head -n1 | tr -d ' ' )
+    TEST_EXPECTED='\a/\r1OK'
+    check "$TEST_RESULT" "$TEST_EXPECTED" .
+    TEST_RESULT=$( printf '\\00000000 a/\\n2\n' | $rhash -c --brief - 2>&1 | head -n1 | tr -d ' ' )
+    TEST_EXPECTED='\a/\n2OK'
+    check "$TEST_RESULT" "$TEST_EXPECTED" .
+  fi
+fi
 TEST_RESULT=$( $rhash -p '\63\1\277\x0f\x1\t\\ \x34\r' -m "" )
 TEST_EXPECTED=$( printf '\63\1\277\17\1\t\\ 4\r' )
 check "$TEST_RESULT" "$TEST_EXPECTED"
@@ -237,7 +265,7 @@ check "$TEST_RESULT" "$TEST_EXPECTED"
 new_test "test file lists:            "
 F="$RHASH_TMP/t"
 touch ${F}1 ${F}2 ${F}3 ${F}4
-( echo ${F}2; echo ${F}3 ) > ${F}l
+printf "${F}2\n${F}3\n" > ${F}l
 TEST_RESULT=$($rhash -p '%f ' ${F}1 --file-list ${F}l ${F}4)
 check "$TEST_RESULT" "t1 t2 t3 t4 "
 rm -f ${F}1 ${F}2 ${F}3 ${F}4 ${F}l
@@ -328,10 +356,27 @@ TEST_RESULT=$( $rhash -vc t.sum 2>&1 | grep 'OK' )
 check "$TEST_RESULT" ""
 rm t.sum
 
+new_test "test missig files:          "
+rm -f a.txt b.txt
+printf "00000000 a.txt\\n00000000 test-empty.file\\n00000000 b.txt" > c.sfv
+TEST_RESULT=$( $rhash --missing c.sfv 2>&1 | tr -d '\r' | tr '\n' '@' )
+check "$TEST_RESULT" "a.txt@b.txt@"
+
+new_test "test unverified files:      "
+rm -rf d/ && mkdir d
+touch d.txt d/a.txt
+printf "00000000 d.txt\\n00000000 d/b.txt" > d.sfv
+TEST_RESULT=$( $rhash -r --unverified d.sfv d.txt test-empty.file d/ 2>&1 | tr -d '\r' | tr '\n' '@' )
+match "$TEST_RESULT" "^test-empty.file@d.a.txt@\$"
+
+new_test "test update:                "
+TEST_RESULT=$( $rhash -r --simple --update d.sfv d.txt test-empty.file d/ 2>&1 | tr -d '\r' | tr '\n' '@' )
+check "$TEST_RESULT" "Updated: d.sfv@"
+
 new_test "test *accept options:       "
 mkdir -p test_dir/a && touch test_dir/a/file.txt test_dir/a/file.bin
-# correctly handle MIGW posix path conversion
-echo "$MSYSTEM" | grep -q '^MINGW[36][24]' && SLASH=// || SLASH="/"
+# correctly handle MINGW posix path conversion
+mingw_or_ucrt && SLASH=// || SLASH="/"
 # test also --path-separator option
 TEST_RESULT=$( $rhash -rC --simple --accept=.bin --path-separator=$SLASH test_dir )
 check "$TEST_RESULT" "00000000  test_dir/a/file.bin" .
@@ -379,7 +424,7 @@ rm -f "$UNWRITABLE_FILE"
 
 # check if any test failed
 if [ $fail_cnt -gt 0 ]; then
-  echo "Failed $fail_cnt checks"
+  printf "Failed $fail_cnt checks\n"
   exit 1 # some tests failed
 fi
 
